import numpy as np
from scipy import stats


def correlation_coefficient(x: np.ndarray, y: np.ndarray) -> tuple[float, float]:
    """
    Считает коэффициент корреляции Пирсона и p-значение.

    Коэффициент корреляции Пирсона измеряет линейную зависимость между двумя наборами данных.
    Изменяется в диапазоне от -1 до +1, где 0 означает отсутствие корреляции.
    Корреляция, равная -1 или +1, означает точную линейную зависимость.
    Положительная корреляция означает, что по мере увеличения x увеличивается и y.
    Отрицательная корреляция означает, что по мере увеличения x уменьшается и y.

    p-значение - это вероятность того, что некоррелированная система создаст наборы данных,
    корреляция Пирсона в которых будет как минимум такой же сильной, как в этих наборах данных.

    Parameters
    ----------
    x: np.ndarray
        1D массив значений.
    y: np.ndarray
        1D массив значений.

    Returns
    -------
    tuple[float, float]
        Коэффициент корреляции и вероятность получения наблюдаемых результатов.
    """

    assert x.ndim == 1, "x: Ожидался 1D массив"
    assert y.ndim == 1, "y: Ожидался 1D массив"
    assert x.shape == y.shape, "x и y должны иметь одинаковую длину"

    result = stats.pearsonr(x, y)
    corr_coef = result.statistic
    p_value = result.pvalue
    return corr_coef, p_value


def spearman_correlation_coefficient(
    x: np.ndarray, y: np.ndarray
) -> tuple[float, float]:
    """
    Считает коэффициент корреляции Спирмена и p-значение.

    Коэффициент корреляции Спирмена — это непараметрическая мера монотонности взаимосвязи между двумя наборами данных.
    Изменяется в диапазоне от -1 до +1, где 0 означает отсутствие корреляции.
    Корреляция, равная -1 или +1, означает точную монотонную взаимосвязь.
    Положительная корреляция означает, что по мере увеличения x увеличивается и y.
    Отрицательная корреляция означает, что по мере увеличения x уменьшается и y.

    p-значение - это вероятность того, что некоррелированная система создаст наборы данных,
    корреляция Спирмена в которых будет как минимум такой же сильной, как в этих наборах данных.

    Parameters
    ----------
    x: np.ndarray
        1D массив значений.
    y: np.ndarray
        1D массив значений.

    Returns
    -------
    tuple[float, float]
        Коэффициент корреляции и вероятность получения наблюдаемых результатов.
    """

    assert x.ndim == 1, "x: Ожидался 1D массив"
    assert y.ndim == 1, "y: Ожидался 1D массив"
    assert x.shape == y.shape, "x и y должны иметь одинаковую длину"

    result = stats.spearmanr(x, y)
    corr_coef = result.statistic
    p_value = result.pvalue
    return corr_coef, p_value


def correlation_coefficients_matrix(data: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
    """
    Считает матрицу корреляции Пирсона и матрицу p-значений.

    Коэффициент корреляции Пирсона измеряет линейную зависимость между двумя наборами данных.
    Изменяется в диапазоне от -1 до +1, где 0 означает отсутствие корреляции.
    Корреляция, равная -1 или +1, означает точную линейную зависимость.
    Положительная корреляция означает, что по мере увеличения x увеличивается и y.
    Отрицательная корреляция означает, что по мере увеличения x уменьшается и y.

    p-значение - это вероятность того, что некоррелированная система создаст наборы данных,
    корреляция Пирсона в которых будет как минимум такой же сильной, как в этих наборах данных.

    Parameters
    ----------
    data: np.ndarray
        Массив значений, размера (N, M).

    Returns
    -------
    tuple[np.ndarray, np.ndarray]
        Матрица корреляции(размера (N, N)) и матрица вероятностей получения наблюдаемых результатов(размера (N, N)).
    """

    assert data.ndim == 2, "data: Ожидался 2D массив"
    assert data.shape[1] >= 2, "data: Ожидался массив с элементами длинной не менее 2"

    matrix = np.zeros((data.shape[0], data.shape[0]), dtype=float)
    p_matrix = np.zeros((data.shape[0], data.shape[0]), dtype=float)
    for i in range(matrix.shape[0]):
        for j in range(matrix.shape[0]):
            matrix[i, j], p_matrix[i, j] = correlation_coefficient(data[i], data[j])
    return matrix, p_matrix


def spearman_correlation_coefficients_matrix(
    data: np.ndarray,
) -> tuple[np.ndarray, np.ndarray]:
    """
    Считает матрицу корреляции Спирмена и матрицу p-значений.

    Коэффициент корреляции Спирмена — это непараметрическая мера монотонности взаимосвязи между двумя наборами данных.
    Изменяется в диапазоне от -1 до +1, где 0 означает отсутствие корреляции.
    Корреляция, равная -1 или +1, означает точную монотонную взаимосвязь.
    Положительная корреляция означает, что по мере увеличения x увеличивается и y.
    Отрицательная корреляция означает, что по мере увеличения x уменьшается и y.

    p-значение - это вероятность того, что некоррелированная система создаст наборы данных,
    корреляция Спирмена в которых будет как минимум такой же сильной, как в этих наборах данных.

    Parameters
    ----------
    data: np.ndarray
        Массив значений, размера (N, M).

    Returns
    -------
    tuple[np.ndarray, np.ndarray]
        Матрица корреляции(размера (M, M)) и матрица вероятностей получения наблюдаемых результатов(размера (M, M)).
    """

    assert data.ndim == 2, "data: Ожидался 2D массив"
    assert data.shape[1] >= 2, "data: Ожидался массив с элементами длинной не менее 2"

    matrix = np.zeros((data.shape[0], data.shape[0]), dtype=float)
    p_matrix = np.zeros((data.shape[0], data.shape[0]), dtype=float)
    for i in range(matrix.shape[0]):
        for j in range(matrix.shape[0]):
            matrix[i, j], p_matrix[i, j] = spearman_correlation_coefficient(
                data[i], data[j]
            )
    return matrix, p_matrix


def confusion_matrix(
    true_labels: np.ndarray, predict_labels: np.ndarray
) -> tuple[np.ndarray, np.ndarray]:
    """
    Считает матрицу ошибок(несоответствий).

    Матрица ошибок — матрица, в которой отображается количество предсказанных и фактических классов.
    Строки представляют истинные классы, а столбцы представляют предсказанные классы.
    Если uniq_labels - это массив уникальных классов, то значение элемента (i, j) в матрице ошибок - это количество раз,
    когда предсказан класс uniq_labels[j], а истинное значение класса uniq_labels[i].

    Parameters
    ----------
    true_labels: np.ndarray
        1D массив истинных классов, длины M.
    predict_labels: np.ndarray
        1D массив предсказанных классов, длины M.

    Returns
    -------
    tuple[np.ndarray, np.ndarray]
        Матрица ошибок, размера (M, M).
        Массив уникальных классов, длины M.
    """

    assert true_labels.ndim == 1, "true_labels: Ожидался 1D массив"
    assert predict_labels.ndim == 1, "predict_labels: Ожидался 1D массив"
    assert (
        true_labels.shape == predict_labels.shape
    ), "true_labels и predict_labels должны иметь одинаковую длину"

    uniq_labels = np.unique(np.hstack((true_labels, predict_labels)))
    matrix = np.zeros((len(uniq_labels), len(uniq_labels)))
    for i in range(len(true_labels)):
        first_idx = int(np.where(uniq_labels == true_labels[i])[0][0])
        second_idx = int(np.where(uniq_labels == predict_labels[i])[0][0])
        matrix[first_idx][second_idx] += 1
    return matrix.astype(int), uniq_labels



from abc import ABC, abstractmethod

import numpy as np
from typing_extensions import Optional


class Calculate(ABC):
    @staticmethod
    @abstractmethod
    def calculate(values: np.ndarray) -> float:
        pass


class Average(Calculate):
    """Класс для вычисления среднего значения в массиве"""

    @staticmethod
    def calculate(values: np.ndarray) -> float:
        """
        Возвращает среднее значение в массиве.

        Parameters
        ----------
        values: np.ndarray
            Массив значений.

        Returns
        -------
        float
            Среднее значение в массиве.
        """

        return float(np.mean(values))


class Min(Calculate):
    """Класс для вычисления минимального значения в массиве"""

    @staticmethod
    def calculate(values: np.ndarray) -> float:
        """
        Возвращает минимальное значение в массиве.

        Parameters
        ----------
        values: np.ndarray
            Массив значений.

        Returns
        -------
        float
            Минимальное значение в массиве.
        """

        return float(np.min(values))


class Max(Calculate):
    """Класс для вычисления максимального значения в массиве"""

    @staticmethod
    def calculate(values: np.ndarray) -> float:
        """
        Возвращает максимальное значение в массиве.

        Parameters
        ----------
        values: np.ndarray
            Массив значений.

        Returns
        -------
        float
            Максимальное значение в массиве.
        """

        return float(np.max(values))


class Median(Calculate):
    """Класс для вычисления медианного значения в массиве"""

    @staticmethod
    def calculate(values: np.ndarray) -> int:
        """
        Возвращает медианное значение в массиве.

        Parameters
        ----------
        values: np.ndarray
            1D массив значений.

        Returns
        -------
        float
            Медианное значение в массиве.
        """
        assert values.ndim == 1, "values: Ожидался 1D массив"

        sorted_values = sorted(values)
        idx = int(len(sorted_values) // 2)
        return sorted_values[idx]


class Top(Calculate):
    """Класс для вычисления первого значения в массиве"""

    @staticmethod
    def calculate(values: np.ndarray) -> float:
        """
        Возвращает первое значение в массиве.

        Parameters
        ----------
        values: np.ndarray
            1D массив значений.

        Returns
        -------
        float
            Первое значение в массиве.
        """
        assert values.ndim == 1, "values: Ожидался 1D массив"

        return float(values[0])


class MostFrequent(Calculate):
    """Класс для вычисления наиболее часто встречаемого значения в массиве"""

    @staticmethod
    def calculate(values: np.ndarray) -> float:
        """
        Возвращает наиболее часто встречаемое значение в массиве.

        Parameters
        ----------
        values: np.ndarray
            Массив значений.

        Returns
        -------
        float
            Наиболее часто встречаемое значение в массиве.
        """

        uniq_values, counts = np.unique(values, return_counts=True)
        return float(uniq_values[np.argmax(counts)])


class LessFrequent(Calculate):
    """Класс для вычисления наименее часто встречаемого значения в массиве"""

    @staticmethod
    def calculate(values: np.ndarray) -> float:
        """
        Возвращает наименее часто встречаемое значение в массиве.

        Parameters
        ----------
        values: np.ndarray
            Массив значений.

        Returns
        -------
        float
            Наименее часто встречаемое значение в массиве.
        """
        uniq_values, counts = np.unique(values, return_counts=True)
        return float(uniq_values[np.argmin(counts)])



import numpy as np
from sklearn.decomposition import PCA

RANDOM_STATE = 42


def get_pca_statistics(
    data: np.ndarray, random_state: int = RANDOM_STATE
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Вычисляет статистику по PCA.

    Parameters
    ----------
    data: np.ndarray
        Массив значений, размера (N, M).
    random_state: int
        Определяет генерацию случайных чисел.

    Returns
    -------
    tuple[np.ndarray, np.ndarray, np.ndarray]
        Массив собственных значений, размера N.
        Массив процентов дисперсии, объясняемых каждым из выбранных компонентов, размера N.
        Массив суммированных процентов дисперсии, размера N.
    """

    assert data.ndim == 2, "data: Ожидался 2D массив"

    pca = PCA(n_components=data.shape[1], random_state=random_state)
    pca.fit(data)
    explained_variance_ratio = pca.explained_variance_ratio_
    sum_explained_variance_ratio = explained_variance_ratio.copy()
    for i in range(1, len(explained_variance_ratio)):
        sum_explained_variance_ratio[i] += sum_explained_variance_ratio[i - 1]
    return pca.singular_values_, explained_variance_ratio, sum_explained_variance_ratio


def pca_transform(
    data: np.ndarray, n_components: int, random_state: int = RANDOM_STATE
) -> np.ndarray:
    """
    Понижение размерности с помощью PCA путём проецирования данных на главные (собственные) вектора.

    Parameters
    ----------
    data: np.ndarray
        Массив значений, размера (N, M).
    n_components: int
        Количество сохраняемых компонентов.
    random_state: int
        Определяет генерацию случайных чисел.

    Returns
    -------
    np.ndarray
        Преобразованный массив, размера (N, n_components).
    """

    assert data.ndim == 2, "data: Ожидался 2D массив"
    assert (
        n_components > 0
    ), "Параметр n_components должен принимать положительное значение"

    return PCA(n_components=n_components, random_state=random_state).fit_transform(data)



from dataclasses import asdict
from typing import List, Tuple

import numpy as np
import numpy.typing as npt
from cratonapi import datacontainers
from cratonapi.dataconnector import DataConnector
from cratonapi.datacontainers import TransformMatrix


class Cube:
    """
    Класс для работы с сейсмическим кубом.

    Attributes
    ----------
    connection : cratonapi.dataconnector.DataConnector
        Объект класса для взаимодействия с WSeis.
    """

    def __init__(self, connection: DataConnector):
        self.connection = connection

    def read(
        self,
        cube_id: int,
        horizon_top_name: str,
        horizon_bot_name: str,
        top_off: float,
        bot_off: float,
        start_inline_idx: int,
        end_inline_idx: int,
        start_crossline_idx: int,
        end_crossline_idx: int,
    ) -> datacontainers.CubeDataSlice:
        """
        Подгружает данные трасс куба в диапазоне.

        Parameters
        ----------
        cube_id: int
            Идентификатор куба.
        horizon_top_name: str
            Имя верхнего горизонта границы.
        horizon_bot_name: str
            Имя нижнего горизонта границы.
        top_off: float
            Смещение для верхнего горизонта.
        bot_off: float
            Смещение для нижней горизонта.
        start_inline_idx: int
            Индекс инлайна с которого начинать набор данных.
        end_inline_idx: int
            Индекс инлайна которым заканчивать набор данных.
        start_crossline_idx: int
            Индекс кросслайна с которого начинать набор данных.
        end_crossline_idx: int
            Индекс кросслайна которым заканчивать набор данных.

        Returns
        -------
        cratonapi.datacontainers.CubeDataSlice
            Куб.
        """

        cube = self.connection.get_cube_data_range(
            cube_id=cube_id,
            horizon_top_name=horizon_top_name,
            horizon_bot_name=horizon_bot_name,
            top_off=top_off,
            bot_off=bot_off,
            start_inline_idx=start_inline_idx,
            end_inline_idx=end_inline_idx,
            start_crossline_idx=start_crossline_idx,
            end_crossline_idx=end_crossline_idx,
        )
        return cube

    def get_info(self) -> dict:
        """
        Подгружает информацию о кубах.

        Returns
        -------
        dict
            Словарь, где ключ - имя куба.
            Значения:
                id: int
                    Идентификатор куба.
                x_min_inl_min_xl: float
                    X координата трассы с минимальными инлайном и кросслайном в кубе.
                y_min_inl_min_xl: float
                    Y координата трассы с минимальными инлайном и кросслайном в кубе.
                x_max_inl_min_xl: float
                    X координата трассы с максимальным инлайном и минимальным кросслайном в кубе.
                y_max_inl_min_xl: float
                    Y координата трассы с максимальным инлайном и минимальным кросслайном в кубе.
                x_min_inl_max_xl: float
                    X координата трассы с минимальным инлайном и максимальным кросслайном в кубе.
                y_min_inl_max_xl: float
                    Y координата трассы с минимальным инлайном и максимальным кросслайном в кубе.
                x_max_inl_max_xl: float
                    X координата трассы с максимальными инлайном и кросслайном в кубе.
                y_max_inl_max_xl: float
                    Y координата трассы с максимальными инлайном и кросслайном в кубе.
                inline_count: int
                    Количество инлайнов в кубе.
                xline_count: int
                    Количество кросслайнов в кубе.
                samples_count: int
                    Количество отсчётов в каждой трассе куба.
                dt: int
                    Шаг дискретизации трасс куба.
                cube_type: int
                    Тип куба.
                min_idx_inline: int
                    Минимальный номер инлайна.
                min_idx_xline: int
                    Минимальный номер кросслайна.
                horizons: list[str]
                    Список имен кубов.
                transform_matrix: cratonapi.datacontainers.TransformMatrix
                    Матрица трансформации.
        """

        cube_list = self.connection.get_cubes()
        info = {}
        for cube in cube_list:
            id_dict = {"id": cube.cube_id}
            properties_dict = asdict(self.connection.get_cube_properties(cube.cube_id))
            horizons_dict = {
                "horizons": list(self.connection.get_cube_horizons(cube.cube_id))
            }
            transforms_dict = {
                "transform_matrix": self.connection.get_transform_matrix(cube.cube_id)
            }

            info_dict = dict(
                id_dict, **properties_dict, **horizons_dict, **transforms_dict
            )
            info[cube.cube_name] = info_dict
        return info

    @staticmethod
    def __get_distance(coord_1: List[float], coord_2: List[float]) -> float:
        return np.sqrt((coord_1[0] - coord_2[0]) ** 2 + (coord_1[1] - coord_2[1]) ** 2)

    @staticmethod
    def __get_transform(
        transform_matrix: TransformMatrix, x: float, y: float
    ) -> Tuple[float, float]:
        x_ = (transform_matrix.a * x + transform_matrix.b * y + transform_matrix.c) / (
            transform_matrix.g * x + transform_matrix.h * y + 1
        )
        y_ = (transform_matrix.d * x + transform_matrix.e * y + transform_matrix.f) / (
            transform_matrix.g * x + transform_matrix.h * y + 1
        )
        return x_, y_

    @staticmethod
    def parse_to_numpy(
        cube: datacontainers.CubeDataSlice, cube_props: dict
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray, float]:
        """
        Преобразовывает куб в numpy массив.

        Parameters
        ----------
        cube: cratonapi.datacontainers.CubeDataSlice
            Куб.
        cube_props: dict
            Словарь со значениями:
                x_min_inl_min_xl: float
                    X координата трассы с минимальными инлайном и кросслайном в кубе.
                y_min_inl_min_xl: float
                    Y координата трассы с минимальными инлайном и кросслайном в кубе.
                x_max_inl_min_xl: float
                    X координата трассы с максимальным инлайном и минимальным кросслайном в кубе.
                y_max_inl_min_xl: float
                    Y координата трассы с максимальным инлайном и минимальным кросслайном в кубе.
                x_min_inl_max_xl: float
                    X координата трассы с минимальным инлайном и максимальным кросслайном в кубе.
                y_min_inl_max_xl: float
                    Y координата трассы с минимальным инлайном и максимальным кросслайном в кубе.
                x_max_inl_max_xl: float
                    X координата трассы с максимальными инлайном и кросслайном в кубе.
                y_max_inl_max_xl: float
                    Y координата трассы с максимальными инлайном и кросслайном в кубе.
                inline_count: int
                    Количество инлайнов в кубе.
                xline_count: int
                    Количество кросслайнов в кубе.
                transform_matrix: cratonapi.datacontainers.TransformMatrix
                    Матрица трансформации.

        Returns
        -------
        tuple[np.ndarray, np.ndarray, np.ndarray, float]
            2D массив трасс куба.
            Сетка координат по x в виде 2D массива.
            Сетка координат по y в виде 2D массива.
            Код бланковки.
        """

        inlines = cube.inlines
        xlines = cube.xlines

        x_cube_coords = np.array(
            [
                cube_props["x_min_inl_min_xl"],
                cube_props["x_max_inl_min_xl"],
                cube_props["x_min_inl_max_xl"],
                cube_props["x_max_inl_max_xl"],
            ]
        )

        y_cube_coords = np.array(
            [
                cube_props["y_min_inl_min_xl"],
                cube_props["y_max_inl_min_xl"],
                cube_props["y_min_inl_max_xl"],
                cube_props["y_max_inl_max_xl"],
            ]
        )

        data = cube.data
        blank_code = np.max(abs(data))
        blank_mask = data == blank_code

        cube_data = [data[i, ~blank_mask[i, :]] for i in range(len(blank_mask))]
        shapes = np.array([cube_data[i].shape[0] for i in range(len(cube_data))])
        min_size = np.min(shapes)
        if min_size == 0:
            idx = np.where(shapes == 0)[0]
            for i in idx:
                if i > 0:
                    cube_data[i] = cube_data[i - 1]
                else:
                    cube_data[i] = cube_data[i + 1]
                shapes[i] = cube_data[i].shape[0]
            min_size = np.min(shapes)
        cube_slice = np.stack([cube_data[i][:min_size] for i in range(len(cube_data))])

        dx = (
            Cube.__get_distance(
                [x_cube_coords[0], y_cube_coords[0]],
                [x_cube_coords[1], y_cube_coords[1]],
            )
            / cube_props["inline_count"]
        )
        dy = (
            Cube.__get_distance(
                [x_cube_coords[0], y_cube_coords[0]],
                [x_cube_coords[2], y_cube_coords[2]],
            )
            / cube_props["xline_count"]
        )

        dx, dy = int(dx), int(dy)

        idx_dict = {
            (int(inlines[i]), int(xlines[i])): i for i in range(inlines.shape[0])
        }

        x_min = np.min(x_cube_coords)
        y_min = np.min(y_cube_coords)

        x_max = np.max(x_cube_coords)
        y_max = np.max(y_cube_coords)

        size = (int((x_max - x_min) / dx), int((y_max - y_min) / dy))
        grid_values = np.zeros((size[1], size[0], len(cube_slice[0])))
        for i in range(size[1]):
            y_i = y_min + i * dy
            for j in range(size[0]):
                x_j = x_min + j * dx
                inl, xl = Cube.__get_transform(cube_props["transform_matrix"], x_j, y_i)
                inl = int(inl * cube_props["inline_count"])
                xl = int(xl * cube_props["xline_count"])
                if (inl, xl) in idx_dict:
                    idx = int(idx_dict[(inl, xl)])  # type: ignore[assignment]
                    grid_values[i, j, :] = cube_slice[idx][:]
                else:
                    grid_values[i, j, :] = np.nan
        grid_values = np.asarray(grid_values)
        grid_values = np.reshape(
            grid_values,
            newshape=(
                grid_values.shape[0] * grid_values.shape[1],
                grid_values.shape[2],
            ),
        )

        x = np.linspace(x_min, x_max, size[0])
        y = np.linspace(y_min, y_max, size[1])

        xx, yy = np.meshgrid(x, y)

        return grid_values, xx, yy, blank_code



import numpy as np
from sklearn import linear_model

RANDOM_STATE = 42


class RANSACRegression:
    """Класс для RANdom SAmple Consensus регрессии"""

    @staticmethod
    def get_model(
        x_train: np.ndarray, y_train: np.ndarray, random_state: int = RANDOM_STATE
    ) -> linear_model.RANSACRegressor:
        """
        Возвращает модель обученную на тренировочных данных.

        Parameters
        ----------
        x_train: np.ndarray
            Массив значений размера (n_samples, n_features).
            Где n_samples - количество образцов, n_features - количество признаков.
        y_train: np.ndarray
            Массив значений размера (n_samples, n_targets) или (n_samples,).
            Где n_samples - количество образцов, n_targets - количество целевых переменных.
        random_state: int
            Определяет генерацию случайных чисел.

        Returns
        -------
        RANSACRegressor
            Обученная модель.
        """

        assert x_train.ndim == 2, "x_train: Ожидался 2D массив"
        assert (
            y_train.ndim == 1 or y_train.ndim == 2
        ), "y_train: Ожидался 1D или 2D массив"
        assert (
            x_train.shape[0] == y_train.shape[0]
        ), "Длина массивов x_train и y_train должна быть равна"

        min_samples = None
        if len(x_train) <= 2:
            min_samples = len(x_train)
        model = linear_model.RANSACRegressor(
            random_state=random_state, min_samples=min_samples
        )
        model.fit(x_train, y_train)
        return model

    @staticmethod
    def predict(model: linear_model.RANSACRegressor, x_test: np.ndarray) -> np.ndarray:
        """
         Прогнозирует значения для тестового набора данных.

        Parameters
        ----------
        model: RANSACRegressor
            Обученная модель.
        x_test: np.ndarray
            Массив значений размера (n_samples, n_features).
            Где n_samples - количество образцов, n_features - количество признаков.

        Returns
        -------
        np.ndarray
            Массив предсказанных значений размера (n_samples, n_targets) или (n_samples,). Где n_targets - количество целевых переменных.
        """

        assert x_test.ndim == 2, "x_test: Ожидался 2D массив"

        prediction = model.predict(x_test)
        return prediction

    @staticmethod
    def get_coeffs(
        model: linear_model.RANSACRegressor,
    ) -> tuple[np.ndarray, float | np.ndarray]:
        """
         Получает коэффициенты линейной регрессии.

        Parameters
        ----------
        model: RANSACRegressor
            Обученная модель.

        Returns
        -------
        tuple[np.ndarray, float | np.ndarray]
            Массив коэффициентов, отвечающих за наклон. Если n_targets = 1 то размер (n_features,). Иначе размер(n_targets, n_features).
                (n_features - количество признаков, n_targets - количество целевых переменных)
            Массив коэффициентов или коэффициент, отвечающий за смещение. Если n_targets = 1 то float. Иначе размер массива (n_targets,).
        """

        a = model.estimator_.coef_
        b = model.estimator_.intercept_
        return a, b


class LinearRegression:
    """Класс для линейной регрессии"""

    @staticmethod
    def get_model(
        x_train: np.ndarray, y_train: np.ndarray
    ) -> linear_model.LinearRegression:
        """
        Возвращает модель обученную на тренировочных данных.

        Parameters
        ----------
        x_train: np.ndarray
            Массив значений размера (n_samples, n_features).
            Где n_samples - количество образцов, n_features - количество признаков.
        y_train: np.ndarray
            Массив значений размера (n_samples, n_targets) или (n_samples,).
            Где n_samples - количество образцов, n_targets - количество целевых переменных.

        Returns
        -------
        LinearRegression
            Обученная модель.
        """

        assert x_train.ndim == 2, "x_train: Ожидался 2D массив"
        assert (
            y_train.ndim == 1 or y_train.ndim == 2
        ), "y_train: Ожидался 1D или 2D массив"
        assert (
            x_train.shape[0] == y_train.shape[0]
        ), "Длина массивов x_train и y_train должна быть равна"

        model = linear_model.LinearRegression()
        model.fit(x_train, y_train)
        return model

    @staticmethod
    def predict(model: linear_model.LinearRegression, x_test: np.ndarray) -> np.ndarray:
        """
         Прогнозирует значения для тестового набора данных.

        Parameters
        ----------
        model: RANSACRegressor
            Обученная модель.
        x_test: np.ndarray
            Массив значений размера (n_samples, n_features).
            Где n_samples - количество образцов, n_features - количество признаков.

        Returns
        -------
        np.ndarray
            Массив предсказанных значений размера (n_samples,).
        """

        assert x_test.ndim == 2, "x_test: Ожидался 2D массив"

        prediction = model.predict(x_test)
        return prediction

    @staticmethod
    def get_coeffs(
        model: linear_model.LinearRegression,
    ) -> tuple[np.ndarray, float | np.ndarray]:
        """
         Получает коэффициенты линейной регрессии.

        Parameters
        ----------
        model: LinearRegression
            Обученная модель.

        Returns
        -------
        tuple[np.ndarray, float | np.ndarray]
            Массив коэффициентов, отвечающих за наклон. Если n_targets = 1 то размер (n_features,). Иначе размер(n_targets, n_features).
                (n_features - количество признаков, n_targets - количество целевых переменных)
            Массив коэффициентов или коэффициент, отвечающий за смещение. Если n_targets = 1 то float. Иначе размер массива (n_targets,).
        """

        a = model.coef_
        b = model.intercept_
        return a, b



import numpy as np
from imblearn.over_sampling import SMOTE
from imblearn.under_sampling import TomekLinks
from sklearn.covariance import EllipticEnvelope
from sklearn.ensemble import IsolationForest
from sklearn.neighbors import LocalOutlierFactor
from sklearn.preprocessing import MaxAbsScaler, MinMaxScaler, StandardScaler
from typing_extensions import Optional

RANDOM_STATE = 42


def drop_nan(data: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
    """
    Удаляет строки в которых присутсвует nan.

    Parameters
    ----------
    data: np.ndarray
        1D или 2D массив значений.

    Returns
    -------
    tuple[np.ndarray, np.ndarray]
        Массив без nan.
        Маска(True - в строке есть nan, False - в строке нет nan), в ввиде 1D массива.
    """

    assert data.ndim == 1 or data.ndim == 2, "data: Ожидался 1D или 2D массив"

    if data.ndim == 1:
        mask = np.isnan(data)
    else:
        mask = np.any(np.isnan(data), axis=1)
    return data[~mask], mask


def max_abs_scaler(
    data: np.ndarray, scaler: MaxAbsScaler = None
) -> tuple[np.ndarray, MaxAbsScaler]:
    """
    Масштабирует массив по его максимальному абсолютному значению. f_norm = f / np.max(abs(f), axis=0)

    Parameters
    ----------
    data: np.ndarray
        Массив значений, размером (N, M).
    scaler: MaxAbsScaler, default=None
        Объект класса MaxAbsScaler.

    Returns
    -------
    tuple[np.ndarray, MaxAbsScaler]
        Преобразованный массив, размером (N, M).
        Объект класса MaxAbsScaler.
    """

    assert data.ndim == 2, "data: Ожидался 2D массив"

    if scaler is None:
        scaler = MaxAbsScaler().fit(data)
    return scaler.transform(data), scaler


def min_max_scaler(
    data: np.ndarray, scaler: MinMaxScaler = None
) -> tuple[np.ndarray, MinMaxScaler]:
    """
    Масштабирует массив до диапазона (0,1). f_norm = (f - f.min(axis=0)) / (f.max(axis=0) - f.min(axis=0))

    Parameters
    ----------
    data: np.ndarray
        Массив значений, размером (N, M).
    scaler: MinMaxScaler, default=None
        Объект класса MinMaxScaler.

    Returns
    -------
    tuple[np.ndarray, MinMaxScaler]
        Преобразованный массив, размером (N, M).
        Объект класса MinMaxScaler.
    """

    assert data.ndim == 2, "data: Ожидался 2D массив"

    if scaler is None:
        scaler = MinMaxScaler().fit(data)
    return scaler.transform(data), scaler


def standard_scaler(
    data: np.ndarray, scaler: StandardScaler = None
) -> tuple[np.ndarray, StandardScaler]:
    """
    Масштабирует массив, удалив среднее значение и увеличив дисперсию до единицы.
    f_norm = (f - np.mean(f, axis=0)) / (np.std(f, axis=0))

    Parameters
    ----------
    data: np.ndarray
        Массив значений, размером (N, M).
    scaler: StandardScaler, default=None
        Объект класса StandardScaler.

    Returns
    -------
    tuple[np.ndarray, StandardScaler]
        Преобразованный массив, размером (N, M).
        Объект класса StandardScaler.
    """

    assert data.ndim == 2, "data: Ожидался 2D массив"

    if scaler is None:
        scaler = StandardScaler().fit(data)
    return scaler.transform(data), scaler


def delete_outliers_LOF(
    data: np.ndarray,
    n_neighbors: int = 20,
    algorithm: str = "auto",
    n_jobs: Optional[int] = None,
) -> tuple[np.ndarray, np.ndarray]:
    """
    Избавляет массив значений от выбросов с помощью локального коэффициента выбросов(Local Outlier Factor).

    Local Outlier Factor (LOF) — это алгоритм для выявления аномальных точек в наборе данных.
    Он делает это путем измерения локальной плотности точек вокруг каждой точки данных и сравнения ее с плотностью точек
    вокруг других точек данных.

    Для вычисления локальной плотности точек вокруг каждой точки данных алгоритм LOF использует меру, называемую
    расстоянием достижимости. Расстояние достижимости точки — это мера того, насколько «сложно» достичь этой точки
    из других точек в наборе данных. Для вычисления расстояния достижимости точки алгоритм LOF сначала определяет
    k ближайших соседей точки, где k — указанный пользователем параметр. Затем он вычисляет расстояние между точкой и
    каждым из ее k ближайших соседей. Расстояние достижимости точки затем определяется как максимальное из этих
    k расстояний. Расстояние достижимости используется для расчета локальной плотности достижимости точки, которая
    представляет собой сумму расстояний между точкой и ее k ближайшими соседями, деленную на k.
    Локальная плотность достижимости точки является мерой локальной плотности точек вокруг точки данных.

    Фактор выброса точки рассчитывается как отношение локальной плотности достижимости точки данных к средней локальной
    плотности достижимости ее k ближайших соседей. Высокий фактор выброса указывает на то, что точка с большей
    вероятностью будет выбросом, тогда как низкий фактор выброса указывает на то, что точка с большей вероятностью
    будет нормальной (не выбросом).

    Parameters
    ----------
    data: np.ndarray
        Массив значений, размером (N, M).
    n_neighbors: int, default=20
        Количество соседей(параметр LocalOutlierFactor).
    algorithm: str, default='auto'
        Алгоритм(параметр LocalOutlierFactor).
    n_jobs: int, default=None
        Количество ядер процессора(параметр LocalOutlierFactor).

    Returns
    -------
    tuple[np.ndarray, np.ndarray]
        Массив значений без выбросов, размером (X, M).
        Маска(True - значение осталось, False - значение выкинули), размером N.
    """

    assert data.ndim == 2, "data: Ожидался 2D массив"

    clf = LocalOutlierFactor(
        n_neighbors=n_neighbors, algorithm=algorithm, n_jobs=n_jobs
    ).fit(data)
    new_data = clf.fit_predict(data)
    mask = new_data == 1
    return data[mask], mask


def delete_outliers_IsolationForest(
    data: np.ndarray,
    n_estimators: int = 100,
    max_samples: str = "auto",
    n_jobs: Optional[int] = None,
    random_state: int = RANDOM_STATE,
) -> tuple[np.ndarray, np.ndarray]:
    """
    Избавляет массив значений от выбросов с помощью алгоритма Isolation Forest.

    IsolationForest «изолирует» наблюдения, случайным образом выбирая признак, а затем
    случайным образом выбирая значение разделения между максимальным и минимальным значениями выбранного признака.

    Поскольку рекурсивное разбиение можно представить в виде древовидной структуры, количество разбиений,
    необходимых для выделения выборки, эквивалентно длине пути от корневого узла до конечного узла.

    Эта длина пути, усреднённая по лесу из таких случайных деревьев,
    является показателем нормальности и функцией принятия решений.

    При случайном разбиении на разделы, аномалии имеют заметно более короткие пути.

    Parameters
    ----------
    data: np.ndarray
        Массив значений, размером (N, M).
    n_estimators: int, default=100
        Количество базовых оценщиков(параметр IsolationForest).
    max_samples: str, default='auto'
        Количество выборок(параметр IsolationForest).
    n_jobs: int, default=None
        Количество ядер процессора(параметр IsolationForest).
    random_state: int
        Определяет генерацию случайных чисел.

    Returns
    -------
    tuple[np.ndarray, np.ndarray]
        Массив значений без выбросов, размером (X, M).
        Маска(True - значение осталось, False - значение выкинули), размером N.
    """

    assert data.ndim == 2, "data: Ожидался 2D массив"

    clf = IsolationForest(
        random_state=random_state,
        n_estimators=n_estimators,
        max_samples=max_samples,
        n_jobs=n_jobs,
    ).fit(data)
    new_data = clf.predict(data)
    mask = new_data == 1
    return data[mask], mask


def delete_outliers_EllipticEnvelope(
    data: np.ndarray,
    support_fraction: Optional[float] = None,
    random_state: int = RANDOM_STATE,
) -> tuple[np.ndarray, np.ndarray]:
    """
    Избавляет массив распределенных по гауссу значений от выбросов с помощью EllipticEnvelope.

    Апроксимирует данные многомерным нормальным распределением.
    Чем меньше вероятность, что точка принадлежит распределению, тем больше вероятность, что она аномальная.

    Parameters
    ----------
    data: np.ndarray
        Массив значений, размером (N, M).
    support_fraction: float, default=None
        Доля точек(параметр EllipticEnvelope).
    random_state: int
        Определяет генерацию случайных чисел.

    Returns
    -------
    tuple[np.ndarray, np.ndarray]
        Массив значений без выбросов, размером (X, M).
        Маска(True - значение осталось, False - значение выкинули), размером N.
    """

    assert data.ndim == 2, "data: Ожидался 2D массив"

    clf = EllipticEnvelope(
        random_state=random_state, support_fraction=support_fraction
    ).fit(data)
    new_data = clf.predict(data)
    mask = new_data == 1
    return data[mask], mask


def tomek_links(
    data: np.ndarray,
    targets: np.ndarray,
    sampling_strategy: str = "auto",
    n_jobs: Optional[int] = None,
) -> tuple[np.ndarray, np.ndarray]:
    """
    Выполняет заниженную выборку путем удаления ссылок Томека.

    Алгоритм находит выборки данных из класса большинства,
    которые имеют наименьшее евклидово расстояние с данными класса меньшинства, а затем удаляет их.

    Parameters
    ----------
    data: np.ndarray
        Массив значений, размера (N1, M1).
    targets: np.ndarray
        Массив меток data, размера N1.
    sampling_strategy: str, default='auto'
        Стратегия выборки(параметр TomekLinks).
    n_jobs: Optional[int], default=None
        Количество ядер процессора(параметр TomekLinks).

    Returns
    -------
    tuple[np.ndarray, np.ndarray]
        Массив, содержащий данные с повторной выборкой, размера (N2, M2).
        Массив меток, размера N2.
    """

    assert data.ndim == 2, "data: Ожидался 2D массив"
    assert targets.ndim == 1, "targets: Ожидался 1D массив"
    assert (
        data.shape[0] == targets.shape[0]
    ), "Массив targets имеет несоответствующую длину"

    tl = TomekLinks(sampling_strategy=sampling_strategy, n_jobs=n_jobs)
    data_res, targets_res = tl.fit_resample(data, targets)
    return data_res, targets_res


def SMOTE_(
    data: np.ndarray,
    targets: np.ndarray,
    sampling_strategy: str = "auto",
    k_neighbors: int = 5,
    n_jobs: Optional[int] = None,
    random_state: int = RANDOM_STATE,
) -> tuple[np.ndarray, np.ndarray]:
    """
    Выполненяет избыточную выборку с использованием метода синтетической передискретизации меньшинства.

    SMOTE синтезирует элементы в непосредственной близости от уже существующих в меньшем наборе.

    Алгоритм выбирает элементы, которые близки, проводит линию между ними и
    добавляет новый элемент в точке вдоль этой линии.

    Parameters
    ----------
    data: np.ndarray
        Массив значений, размера (N1, M1).
    targets: np.ndarray
        Массив меток data, размера N1.
    sampling_strategy: str, default='auto'
        Стратегия выборки(параметр SMOTE).
    k_neighbors: int, default=5
        Количество ближайших соседей(параметр SMOTE).
    n_jobs: int, default=None
        Количество ядер процессора(параметр SMOTE).
    random_state: int
        Определяет генерацию случайных чисел.

    Returns
    -------
    tuple[np.ndarray, np.ndarray]
        Массив, содержащий данные с повторной выборкой, размера (N2, M2).
        Массив меток, размера N2.
    """

    assert data.ndim == 2, "data: Ожидался 2D массив"
    assert targets.ndim == 1, "targets: Ожидался 1D массив"
    assert (
        data.shape[0] == targets.shape[0]
    ), "Массив targets имеет несоответствующую длину"

    sm = SMOTE(
        sampling_strategy=sampling_strategy,
        k_neighbors=k_neighbors,
        n_jobs=n_jobs,
        random_state=random_state,
    )
    data_res, targets_res = sm.fit_resample(data, targets)
    return data_res, targets_res



"""Модуль исключений при работе со скважинными данными.
"""


class WellException(BaseException):
    """Базовый класс исключений при работе со скважинными данными"""

    pass


class WellTrajectoryDepthException(WellException):
    """"""

    pass


class WellEmptyCurvesException(WellException):
    """"""

    pass



import numpy as np
from cratonapi import datacontainers
from cratonapi.dataconnector import DataConnector

from cratonml.exceptions.DataExceptions import EmptyInputDataException


class Outline:
    """
    Класс для работы с контурами.

    Attributes
    ----------
    connection : cratonapi.dataconnector.DataConnector
        Объект класса для взаимодействия с WSeis.
    """

    def __init__(self, connection: DataConnector):
        self.connection = connection

    def read(
        self, outline_id_list: list[int]
    ) -> tuple[list[datacontainers.Outline], list[int]]:
        """
        Подгружает контуры.

        Parameters
        ----------
        outline_id_list: list[int]
            Список идентификаторов контуров.

        Returns
        -------
        tuple[list[datacontainers.Outline], list[int]]
            Список контуров.
            Список идентификаторов контуров.
        """

        if len(outline_id_list) == 0:
            raise EmptyInputDataException("Список идентификаторов пустой")

        outline_list = []
        outline_ids = []
        for outline_id in outline_id_list:
            outlines = self.connection.get_outline(outline_id)
            for outline in outlines:
                if outline.coordinates.shape[0] != 0:
                    outline_list.append(outline)
                    outline_ids.append(outline_id)
        return outline_list, outline_ids

    def get_info(self) -> dict:
        """
        Подгружает информацию о контурах.

        Returns
        -------
        dict
            Словарь, где ключ - имя контура.
            Значения:
                id: int
                    Идентификатор контура.
                pen_width: int
                    Ширина линии контура.
                pen_style: int
                    Стиль отображения линии контура(Qt.PenStyle).
                pen_color: tuple[float, float, float, float]
                    Цвет отображения линии контура(Красный, Зелёный, Синий, Альфа). Значения от 0 до 1.
                outline_width: int
                    Ширина подложки контура.
                outline_style: int
                    Стиль отображения подложки контура(Qt.PenStyle).
                outline_color: tuple[float, float, float, float]
                    Цвет отображения подложки контура(Красный, Зелёный, Синий, Альфа). Значения от 0 до 1.
                fill_style: int
                    Стиль отображения заливки(Qt.BrushStyle).
                fill_color: tuple[float, float, float, float]
                    Цвет отображения заливки(Красный, Зелёный, Синий, Альфа). Значения от 0 до 1.
        """

        outlines_list = self.connection.get_outlines_list()
        info = {}
        for outline in outlines_list:
            id_dict = {"id": outline.outline_id}
            pen_width_dict = {"pen_width": outline.pen_width}
            pen_style_dict = {"pen_style": outline.pen_style}
            pen_color_dict = {
                "pen_color": (
                    outline.pen_color.red / 255,
                    outline.pen_color.green / 255,
                    outline.pen_color.blue / 255,
                    outline.pen_color.alpha / 255,
                )
            }
            outline_width_dict = {"outline_width": outline.outline_width}
            outline_style_dict = {"outline_style": outline.outline_style}
            outline_color_dict = {
                "outline_color": (
                    outline.outline_color.red / 255,
                    outline.outline_color.green / 255,
                    outline.outline_color.blue / 255,
                    outline.outline_color.alpha / 255,
                )
            }
            fill_style_dict = {"fill_style": outline.fill_style}
            fill_color_dict = {
                "fill_color": (
                    outline.fill_color.red / 255,
                    outline.fill_color.green / 255,
                    outline.fill_color.blue / 255,
                    outline.fill_color.alpha / 255,
                )
            }

            info_dict = dict(
                **id_dict,
                **pen_width_dict,
                **pen_style_dict,
                **pen_color_dict,
                **outline_width_dict,
                **outline_style_dict,
                **outline_color_dict,
                **fill_style_dict,
                **fill_color_dict
            )
            info[outline.outline_name] = info_dict
        return info

    @staticmethod
    def parse_to_numpy(
        outline_list: list[datacontainers.Outline], dx: float = 25, dy: float = 25
    ) -> tuple[list, list, list]:
        """
        Преобразовывает список контуров в список numpy массивов.

        Parameters
        ----------
        outline_list: list[cratonapi.datacontainers.Outline]
            Список контуров.
        dx: float, default=25
            Шаг по сетке x.
        dy: float, default=25
            Шаг по сетке y.

        Returns
        -------
        tuple[list, list, list]
            Список 1D массивов, каждый из которых является значениями контура.
            Список из сеток координат по x каждого контура в виде 2D массивов.
            Список из сеток координат по y каждого контура в виде 2D массивов.
        """
        assert dx > 0, "dx должен принимать положительное значение"
        assert dy > 0, "dy должен принимать положительное значение"

        outlines = []
        xx = []
        yy = []

        for outline in outline_list:
            x_val, y_val, z_val = outline.coordinates.T
            min_x = np.nanmin(x_val)
            max_x = np.nanmax(x_val)
            min_y = np.nanmin(y_val)
            max_y = np.nanmax(y_val)

            x_line = np.arange(min_x, max_x + dx, dx)
            y_line = np.arange(min_y, max_y + dy, dy)

            grid_xx, grid_yy = np.meshgrid(x_line, y_line)
            map_data = np.zeros(grid_xx.shape)
            map_data[:, :] = np.nan

            uniq_values, counts = np.unique(z_val, return_counts=True)
            uniq_class = np.rint(uniq_values[np.argmax(counts)])
            for i in range(map_data.shape[0]):
                for j in range(map_data.shape[1]):
                    x_ = grid_xx[i, j]
                    y_ = grid_yy[i, j]
                    if np.isnan(map_data[i, j]):
                        k = len(x_val) - 1
                        c = False
                        for u in range(len(x_val)):
                            if (y_val[u] <= y_ < y_val[k]) or (
                                y_val[k] <= y_ < y_val[u]
                            ):
                                if y_val[k] != y_val[u]:
                                    if x_ >= (
                                        (x_val[k] - x_val[u])
                                        * (y_ - y_val[u])
                                        / (y_val[k] - y_val[u])
                                        + x_val[u]
                                    ):
                                        c = not c
                            k = u
                        if c:
                            map_data[i, j] = int(uniq_class)
            outlines.append(map_data.ravel())
            xx.append(grid_xx)
            yy.append(grid_yy)
        return outlines, xx, yy
